         TITLE    8290.6 GST 226  Example Program
*8290.6 GST 226  EXAMPLE PROGRAM
*
*  This program is provided as a rudimentary example of how to write an
*assembler program to run under QDOS.  The following should be noticed:
*
*  1  The equivalences necessary for talking to QDOS and the standard
*     MACRO library are read in with INCLUDE statements but their 
*     contents are not listed because they contain NOLIST directives.
*
*  2  There is a SECTION directive before any code or data is assembled;
*     this is important.
*
*  3  All programs need a program header of the form shown.
*
*  4  Right at the very end there is an END directive.
*
*  5  The use of upper case and lower case in this program is purely a
*     matter of taste.  There is no reason why all the directives and
*     instructions are in upper case other than that the author prefers
*     that particular style.

         PAGE
*First we read in the equivalence and MACRO files.  Note that these are
*not listed in full:  they contain LIST & NOLIST directives.

         INCLUDE  'flp1_QDOS1_IN'         Read in some QDOS equivalences
         INCLUDE  'flp1_MACRO_LIB'        Read in the standard MACROs
*
*For this program the constant SD.CLEAR is required which is in parameter
*file 2 (QDOS2_IN). However if both parameter files are included the
*number of symbols is too large to be able to assemble the file
*Therefore we will add the equivalence here
*
SD.CLEAR EQU   $20

         PAGE
*Now before generating any code or data we must have a SECTION directive.
*It doesn't matter much what the section is called.

         SECTION  CODE

*The very first real thing to be assembled must be the header:
*It has the following form:
*
*  Byte positions    Contents
*
*       0..5         The first instruction to be executed plus filler
*       6..7         A 16-bit word containing the value 4AFB (hexadecimal)
*       8..9         A 16-bit word containing the length of ...
*       A...         ... the name of the program.
*
*It is assembled like this (note the STRING$ MACRO is in MACRO_LIB):

BASE     BRA.S    START             The first real thing to assemble!
         DS.B     6-(*-BASE)        Filler ('*' is where we are)
         DC.W     $4AFB             This is necessary
         STRING$  'EXAMPLE'         This is the name of the program.

*After the header we can have more or less whatever we like.  What appears
*here is the remainder of the essence of a program:

START    BSR.S    MAIN              Do what the program has to do.

*        having done it, finish off as follows:

         MOVEQ    #-1,D1            Negative number to remove THIS job
         CLR.L    D3                Say there were no errors
         QDOSMT$  MT.FRJOB          Force remove job from program area.

*The error exit code is very like the normal exit code except that the
*error number is returned in D3.  Note that it is O.K. to jump to here
*from anywhere:  QDOS can unwind the stack.
*D0 is assumed to hold an error number.

ERROR    MOVE.L   D0,D3             Error number
         MOVEQ    #-1,D1            Negative number to kill THIS job.
         QDOSMT$  MT.FRJOB          Call QDOS to exit with error.

         PAGE
*There follows now the code and data specific to the example program;
*if you are using this source as a basis on which to build a program
*of your own, this is the bit to change to do what your program has to do.

*First a little bit of data...

CH.EOL   EQU      $0A               End of line code

SCREEN   STRING$  'CON_'            QDOS' name for the screen
SIGNON   STRING$  {'This is the example program for the macro assembler.',CH.EOL}
DIRHEAD  STRING$  {'Sorted directory of flp1_',CH.EOL}
DRIVE2   STRING$  'flp1_'

*Finally the code...

MAIN     MOVEQ    #-1,D1            Open for this job ...
         CLR.L    D3                ... exclusively ...
         LEA.L    SCREEN(PC),A0     ... a screen window.
         QDOSOC$  IO.OPEN           QDOS puts channel no. into A0
         TST.W    D0                Look at the return code.
         BNE      ERROR             Go to the error exit if problem.

*We now have a screen channel opened whose channel number is held in A0
*AS there may already be data on the screen it is useful to clear
*the screen so that the data from our program may be read more easily
*
         MOVEQ    #-1,D3            Infinite timeout
         QDOSIO$  SD.CLEAR          clear the screen
         TST.W    D0                Did this succeed?
         BNE      ERROR             If not, take the error exit.

*At this point A0 holds the channel number for the screen;  no other
*register contents are significant (except of course A7).

         LEA.L    SIGNON(PC),A1     Get string address is in A1 ...
         MOVE.W   (A1)+,D2          ... and length is in D2 ...
         QDOSIO$  IO.SSTRG          Output string to the channel in A0.
         TST.W    D0                Did this succeed?
         BNE      ERROR             If not, take the error exit.

*Most programs start something like this but what follows is much more
*specific to this particular program ... so we'll start a new page for it.
         PAGE
*Now the external declarations.  These are for the benefit of the linker.
*Ordinarily these would appear on the first page of any module.

         XREF     COMPARE           Comparison routine
         XREF     SORT              Sort routine

         XDEF     MEMGET            Memory acquisition routine used by SORT
         XDEF     MEMDROP           Memory release routine also used by SORT

*Before going on to the more complicated things, define some more
*equivalences ...

ENTRIES  EQU      50                Largest number of files expected
DIRLEN   EQU      64                Size of a directory record
FNAMLEN  EQU      36                Maximum length of a file name
FNAMBAS  EQU      $0E               position of file name in dir. rec.

*The following equivalences are set up as offsets because they will be
*used as offsets to places in a data area where relevant variables are
*kept.  Note that the OFFSET directive effectively undoes the work of
*the SECTION directive earlier, so we will need another one later.  

         OFFSET   0
SCRNCHN  DS.L     1                 Screen channel
BUF1     DS.L     1                 Base address of one buffer
BUF2     DS.L     1                 Base address of the other buffer
BUF1LEN  DS.L     1                 Size of first buffer
BUF2LEN  DS.L     1                 Size of second buffer

DALEN    DS.B     0                 DALEN = length of this data area

*Also define a MACRO to trace how the program is going.
*It just outputs a given string to the screen.

TRACE$   MACRO    TEXTARG
         IFNUM    [TRACING] = 0 GOTO SKIPTRACE
         PUSH$.L  A1                Preserve register we're going to use
         LEA.L    TRTX[.L],A1       Get address of string to output
         BRA.S    TRCL[.L]          Jump around the string
TRTX[.L] STRING$  {'[TEXTARG]'}
         DC.B     CH.EOL            Terminate string with a newline
TRCL[.L] BSR      TRACE             Call subroutine to do output
         POP$.L   A1                Restore destroyed register
SKIPTRACE MACLAB
         ENDM

TRACING  SETNUM   1                 Tracing on. SETNUM 0 for off.

         SECTION  CODE              No longer in "offset" mode

*Now, to continue with the code, this grabs some memory, scans the
*directory for flp1_, picks the non-zero length filenames out of it
*and lists them on the screen before exiting.

         PUSH$.L  A0                Remember screen channel no.

         MOVEQ    #DALEN,D0         Get size of data area
         BSR      MEMGET            Grab that much memory
         MOVE.L   A0,A5             Keep data area address in A5

         POP$.L   SCRNCHN(A5)       Recover & store channel no.

         TRACE$   {Got data area}

         MOVE.L   #ENTRIES*DIRLEN,D0
         BSR      MEMGET            Grab a buffer
         MOVE.L   A0,BUF1(A5)       Remember where it is
         MOVE.L   D0,BUF1LEN(A5)    and how big it is

         TRACE$   {Got first buffer}

         MOVE.L   #ENTRIES*(FNAMLEN+1),D0
         BSR      MEMGET            Grab another buffer
         MOVE.L   A0,BUF2(A5)       Remember where that is too
         MOVE.L   D0,BUF2LEN(A5)    and how big it is

         TRACE$   {Got second buffer}

*Having grabbed the memory, the next thing to do is to open and
*read the directory

         MOVEQ    #-1,D1            Open for this job ...
         MOVEQ    #4,D3             ... a directory ...
         LEA.L    DRIVE2(PC),A0     ... called "flp1_"
         QDOSOC$  IO.OPEN           Get channel no. into A0

         TRACE$   {Opened directory channel}

         MOVE.L   BUF1(A5),A1       Get buffer start address
         MOVEQ    #DIRLEN,D2        No. of bytes to read each time
         MOVEQ    #-1,D3            "infinite" timeout

         CLR.W    D4                Clear count of directory entries
         REPEAT
            QDOSIO$  IO.FSTRG       Read a record
            ADDQ.W   #1,D4          Count it
            TST.W    D0             Look at the return code
         UNTIL    NE OR D4 GE.W #ENTRIES

         IF       D4 GE.W #ENTRIES THEN.S
            TRACE$   {Too many entries}
            MOVEQ    #ERR.BO,D0     Buffer overflow
         ENDIF
         IF       D0 NE.W #ERR.EF THEN.S
            BRA      ERROR          End of file is normal!
         ENDIF
         PAGE
*Now we have read the entire directory into BUF1.  Next we extract the
*filenames and put them into BUF2, delimited by end-of-line codes.
*N.B.  D4 contains no. of entries, D2 contains the size of an entry
*    & A5 still points to data area.

         MOVE.L   BUF1(A5),A1       Address of directory entries
         MOVE.L   BUF2(A5),A2       Address of buffer for file names
         CLR.L    D5                Initialise count of file names

         FOR.W    D0 = #1 TO D4 DO.S

            LEA.L    FNAMBAS(A1),A3    Filename address in dir. record
            MOVE.W   (A3)+,D3          length of file name

            WHILE    D3 GT.W #0 DO.S
               MOVE.B   (A3)+,(A2)+    Copy the file name
               SUBQ.W   #1,D3
            ENDWHILE

            IF       FNAMBAS(A1) GT.W #0 THEN.S
               MOVE.B   #CH.EOL,(A2)+  Terminate (non-null) file name
               ADDQ.L   #1,D5          Count it
            ENDIF

            ADD.L    D2,A1          Point to next directory record

         ENDFOR

*Release the first buffer

         MOVE.L   BUF1(A5),A0
         BSR.S    MEMDROP
         TRACE$   {First buffer dropped}

*SORT the second buffer

         MOVE.L   D5,D0             Number of records to sort
         MOVEQ    #CH.EOL,D1        Record delimiter
         MOVE.L   BUF2(A5),A0       Address of buffer with data to be sorted
         LEA.L    COMPARE(PC),A1    Start address of comparison routine
         CALL$    SORT

*Print a header for the sorted directory

         MOVE.L   SCRNCHN(A5),A0    Screen channel number
         LEA.L    DIRHEAD(PC),A1    Sorted directory header
         MOVE.W   (A1)+,D2          Buffer start and length
         QDOSIO$  IO.SSTRG          Send string to screen

*Finally write the second buffer to the screen.

         MOVE.L   A2,D2             From last byte used pointer
         SUB.L    BUF2(A5),D2       Calculate no. of used bytes
         MOVE.L   #-1,D3            "infinite" timeout
         MOVE.L   SCRNCHN(A5),A0    Screen channel number
         MOVE.L   BUF2(A5),A1       Buffer base address
         QDOSIO$  IO.SSTRG          Send string to screen

         RTS                        Return to top level
         PAGE
*Subroutine grab a piece of memory from the common heap.
*It aborts the program if the memory cannot be got.

*   Register       On entry                   On exit

*   D0             No. of bytes required      No. of bytes obtained
*   A0             don't care                 address of grabbed memory
*   others         don't care                 preserved

*For this routine an extra stack handling MACRO is required.
*It gets the top item from the stack without changing the stack pointer.

TOP$     MACRO    A
         MOVE.[.ext] (SP),[A]
         ENDM

         SUBROUTINE$ MEMGET,D1-D3/A1-A3

         PUSH$.L  D0                Save parameter
         QDOSMT$  MT.INF            Find out system info.
         MOVE.L   D1,D2             get job i.d.
         TOP$.L   D1                no. of bytes required
         QDOSMT$  MT.ALCHP          attempt to allocate them
         POP$.L   D2                get no. wanted
         IF     D1 LT.L D2 THEN.S   if less than no. acquired
            MOVEQ    #ERR.OM,D0     use QDOS' out-of-memory error
         ENDIF

         TST.W    D0                if there was a problem ...
         BNE      ERROR             ... abort the job

         MOVE.L   D1,D0             otherwise return the count
         END$     MEMGET

         PAGE
*Subroutine to return a piece of memory to the common heap.
*It aborts if there is an error.

*        Register    On entry                      On exit
*
*        A0.L        address of RAM to release     preserved
*        others      don't care                    preserved

         SUBROUTINE$ MEMDROP,D0-D3/A0-A3

         QDOSMT$     MT.RECHP

         TST.W       D0
         BNE         ERROR

         END$        MEMDROP
         PAGE
*This is the trace subroutine which exists just for use by the TRACE$
*MACRO.  It simply outputs the string addressed by A1 and the following
*byte which is assumed to be a new line code. A5 must point
*to the data area defined above (in which the offset SCRNCHN contains
*the screen channel number.  All registers except A1 are preserved.

         SUBROUTINE$ TRACE,D0-D3/A0/A2/A3

         MOVE.W   (A1)+,D2          Get string length, point at it.
         ADDQ.W   #1,D2             Output next byte too
         MOVEQ    #-1,D3            "infinite" timeout
         MOVE.L   SCRNCHN(A5),A0    Get screen channel

         QDOSIO$  IO.SSTRG          Write string to screen

         TST.W    D0                Look at return code
         BNE      ERROR             Abort if error

         END$     TRACE             Restore registers & return
         PAGE
*The very very last thing every module to be assembled must have is ...

         END                        an END directive (optional comment)
